unit MasterMindEngine;

{ MASTER MIND ENGINE v1.4.2 ----------------------------------------------------------------------
  by f0xi [www.delphifr.com]

  v1.4.2 - 29/07/2006
         - Ajout des evenements OnBeforeNewGame et OnAfterNewGame
         - Suppression de l'evenement OnNewGame pour l'evenement OnAfterNewGame
         - Refact. source code et commentaires

  v1.4.0 - 24/07/2006
         - Ajout de l'evenement OnNewGame
         - Modification du generateur de combinaisons secrete
         - Modification de la procedure NewGame
         - Correction du comparateur combinaisons
         - Refact. source code et commentaires

  v1.3.0 - 23/07/2006
         - Ajout des evements OnWin OnLose
         - Ajout de l'evenemant OnEndGame
         - Modification de la procedure SubmitSolve
         - Correction du generateur de combinaison secrete
         - Correction du comparateur de combinaisons

  v1.0.0 - 10/05/2006

 Utilisation -------------------------------------------------------------------------------------
   nouvelle partie :
     TMasterMindEngine.NewGame( longeur des combinaisons 4..8,
                                nombre de coups 10..20,
                                mode des combinaisons : 1 simple, 2 double, 3 triple ect...
                              );

   abandon de la partie :
     TMasterMindEngine.EndGame;

   soumettre une combinaison :
     TMasterMindEngine.SubmitSolve(combinaison TMMEStr 4..8);

   regler le nombre de plots et leur lettre d'identification (A..Z) :
     TMasterMindEngine.PlotStr := 'ABCDEFGHIJ';
     Appelez EndGame avant la modification et NewGame aprs modification 

   lecture des parametres (proprits en lecture seule) :
     combinaison secrete  : TMasterMindEngine.Secrets;
     solutions joues     : TMasterMindEngine.Solves[index];
     verifications        : TMasterMindEngine.Checks[index];
     nbr de coups a jouer : TMasterMindEngine.SolvesLeft;
     nbr de coups max     : TMasterMindEngine.MaxSolves;
     longeur des combi.   : TMasterMindEngine.MaxPlots;


 Evenements --------------------------------------------------------------------------------------
   TMasterMindEngine.OnCheck
      SubmitSolve();
      cet evenement est declanch aprs chaque appel a SubmitSolve();
      utilisez le pour rafraichir votre interface graphique (plateau de jeux)

   TMasterMindEngine.OnWin
      SubmitSolve();
      cet evenement est declanch si la combinaison secrete est resolue
      utilisez le pour transmettre au joueur le fait qu'il a reussi la partie

   TMasterMindEngine.OnLose
      SubmitSolve();
      cet evenement est declanch si la combinaison secrete n'a pas te resolue
      utilisez le pour transmettre au joueur le fait qu'il a perdus

   TMasterMindEngine.OnBeforeNewGame
      NewGame();
      cet evenement est declanch aprs l'appel a la procedure NewGame, avant l'initialisation
      du moteur du jeux.
      utilisez le pour initialiser des parametres avant l'initialisation du moteur de jeux.

   TMasterMindEngine.OnAfterNewGame
      NewGame();
      cet evenement est declanch aprs l'appel a la procedure NewGame, aprs l'initialisation
      du moteur du jeux.
      utilisez le pour initialiser les parametres du plateau de jeux, scores, variables ect...

   TMasterMindEngine.OnEndGame
      SubmitSolve(); EndGame();
      cet evenement est declanch quand il n'y a plus de coups a jouer ou quand la partie est gagne
      utilisez le pour afficher la combinaison secrete qu'il fallait trouver

 ------------------------------------------------------------------------------------------------- }

interface
uses windows, sysutils, classes;

type
  TMMEStr        = string[8];
  TMMEStrArray   = array[1..20] of TMMEStr;

  TMMECheckEvent = procedure(Sender : TObject; const Solve,Check : TMMEStr; const Win : boolean) of object;

  TMasterMindEngine = class(TComponent)
  private
     fSecret     : TMMEStr;
     fSolves     : TMMEStrArray;
     fChecks     : TMMEStrArray;
     fSolvesLeft : integer;
     fMaxPlots   : byte;
     fMaxSolves  : byte;
     fPLotStr    : string;
     fCheckStr   : string;
     fInGame     : boolean;
     fOnCheck    : TMMECheckEvent;
     fOnWin      : TNotifyEvent;
     fOnLose     : TNotifyEvent;
     fOnBeforeNewGame : TNotifyEvent;
     fOnAfterNewGame  : TNotifyEvent;
     fOnEndGame  : TNotifyEvent;
     function GetMMESTR(index, mode : integer) : TMMEStr;
     procedure ResetArray;
  public
     constructor Create(AOwner : TComponent); override;
     property Secrets    : TMMEStr     read fSecret;
     property Solves[index : integer] : TMMEStr index 0 read GetMMESTR;
     property Checks[index : integer] : TMMEStr index 1 read GetMMESTR;
     property SolvesLeft : integer     read fSolvesLeft;
     property MaxSolves  : byte        read fMaxSolves;
     property MaxPlots   : byte        read fMaxPlots;
     procedure NewGame(const AMaxPlots, AMaxSolves, CombinaisonMode : byte);
     procedure EndGame;
     procedure SubmitSolve(const Solve : TMMEStr);
     function IsInGame : boolean;
  published
     property PlotStr    : string         read fPlotStr    write fPlotStr;
     property CheckStr   : string         read fCheckStr;
     property OnCheck    : TMMECheckEvent read fOnCheck    write fOnCheck;
     property OnWin      : TNotifyEvent   read fOnWin      write fOnWin;
     property OnLose     : TNotifyEvent   read fOnLose     write fOnLose;
     property OnBeforeNewGame : TNotifyEvent   read fOnBeforeNewGame  write fOnBeforeNewGame;
     property OnAfterNewGame  : TNotifyEvent   read fOnAfterNewGame   write fOnAfterNewGame;
     property OnEndGame  : TNotifyEvent   read fOnEndGame  write fOnEndGame;
  end;

function DupeChar(const AChar: Char; ACount: Integer): string;
function CountChar(const AChar : char; const S : string) : integer;

implementation

uses math;

{ ------------------------------------------------------------------------------------------------ }

function DupeChar(const AChar: Char; ACount: Integer): string;
var i : integer;
begin
  SetLength(Result, ACount);
  for i := 1 to ACount do result[i] := AChar;
end;

function CountChar(const AChar : char; const S : string) : integer;
var i : integer;
begin
  result := 0;
  for i := 1 to length(s) do
      if AChar = S[i] then inc(result);
end;

{ TMasterMindEngine ------------------------------------------------------------------------------ }

constructor TMasterMindEngine.Create(AOwner : TComponent);
begin
  inherited create(AOwner);
  fInGame     := false;
  fPlotStr    := 'ABCDEFGHIJ';
  fCheckStr   := '12';
  fSolvesLeft := 0;
  fSecret     := '??????';
  fMaxPlots   := 6;
  fMaxSolves  := 12;
  ResetArray;
end;

{ private }

function TMasterMindEngine.GetMMESTR(index, mode : integer) : TMMEStr;
begin
  case Mode of
    0 : result := fSolves[index];
    1 : result := fChecks[index];
  end;
end;

procedure TMasterMindEngine.ResetArray;
var Dc : TMMEStr;
    i  : integer;
begin
  Dc := DupeChar('?',fMaxPlots);
  for i := 1 to high(fSolves) do begin
      fSolves[i] := Dc;
      fChecks[i] := Dc;
  end;
end;

{ public }

procedure TMasterMindEngine.NewGame(const AMaxPlots, AMaxSolves, CombinaisonMode : byte);
var i : integer;
    r : Char;
begin
  {>> si une partie est en cours on sort }
  if fInGame then exit;

  {>> on mets tout de suite fInGame a true pour eviter les doubles appel a NewGame }
  fInGame     := true;

  {>> on initialise les premieres proprits du moteur }
  fMaxPlots   := AMaxPlots;
  fMaxSolves  := 12;
  fSolvesLeft := AMaxSolves;

  {>> on declanche l'evenement OnBeforeNewGame }
  if assigned(fOnBeforeNewGame) then
     fOnBeforeNewGame(self);

  {>> on initialise la combinaison secrete }
  { on ajuste la taille de fSecret }
  SetLength(fSecret,fMaxPlots);
  { on initialise le generateur de nombres aleatoire }
  Randomize;
  { on genere quelques nombres dans le vide pour eviter les repetitions }
  for i := RandomRange(1,11) to RandomRange(20,31) do Random(666);
  { on genere enfin la combinaison secrete }
  for i := 1 to fMaxPlots do begin
      { on tire un pion au hasard }
      r := PlotStr[randomrange(1,Length(PlotStr)+1)];
      { si on empeche pas les doublons, tripls et quadrupls }
      if CombinaisonMode > 1 then begin
         { la condition est ass particuliere puisqu'on compte par rapport a la limite
           grace a countchar, donc si le pion existe deja on ne le corrige pas
           tant que countchar n'est pas egal a CombinaisonMode }
         while (pos(r,fSecret) <> 0) and (countChar(r,fSecret) = CombinaisonMode) do
               r := PlotStr[randomrange(1,Length(PlotStr)+1)];
      end else
      { si on veux des combinaison a pion unique (combinaisonmode = 1) }
         { tant qu'on tire un pion existant on regenere un pion }
         while (pos(r,fSecret) <> 0) do
               r := PlotStr[randomrange(1,Length(PlotStr)+1)];
      { enfin on assigne le pion a l'indice en cours dans la combinaison secrete }
      fSecret[i] := r;
  end;

  {>> on reinitialise les tableaux (Solves et Checks) }
  ResetArray;

  {>> on declanche l'evement OnAfterNewGame }
  if assigned(fOnAfterNewGame) then
     fOnAfterNewGame(self);
end;


procedure TMasterMindEngine.EndGame;
begin
  {>> si on est pas en cours de partie on sort }
  if not fInGame then exit;

  {>> on place fInGame a false }
  fInGame := false;

  {>> on declanche l'evenement fOnEndGame }
  if assigned(fOnEndGame) then
     fOnEndGame(self);
end;


procedure TMasterMindEngine.SubmitSolve(const Solve : TMMEStr);
var i,idx : integer;
    win   : boolean;
    schk  : string;
begin
  if not fInGame then exit;
  if pos('?',Solve) <> 0 then exit;
  if length(Solve) <> Self.MaxPlots then exit;

  if fSolvesLeft = 0 then EndGame;

  win := true;
  dec(fSolvesLeft);
  idx := fMaxSolves - fSolvesLeft;
  fSolves[idx] := Solve;
  schk := '';
  fChecks[idx] := dupechar('0',fMaxPlots);

  for i := 1 to fMaxPlots do
      if Solve[i] = fSecret[i] then begin
         fChecks[idx][i] := '1';
         schk := schk + Solve[i];
      end;

  for i := 1 to fMaxPlots do
      if Solve[i] = fSecret[i] then begin
         fChecks[idx][i] := '1';
      end else
      if pos(Solve[i],fsecret) <> 0 then begin
         if countchar(solve[i],fsecret) > countchar(solve[i],schk) then begin
            fChecks[idx][i] := '2';
            schk := schk + solve[i];
         end;
      end;

  for i := 1 to fMaxPlots do
      win := win and (fChecks[idx][i] = '1');

  if assigned(fOnCheck) then fOnCheck(self,fSolves[idx],fChecks[idx], Win);

  if win then begin
     fSolvesLeft := 0;
     if assigned(fOnWin) then fOnWin(self);
  end;

  if (fSolvesLeft = 0) then begin
     if (not Win) and assigned(fOnLose) then fOnLose(self);
     if assigned(fOnEndGame) then fOnEndGame(self);
     fInGame := false;
  end;
end;

function TMasterMindEngine.IsInGame : boolean;
begin
  result := fInGame;
end;

end.
 